@@ -5,6 +5,11 @@ import torch
 from functools import wraps
 from einops import rearrange, repeat
 
+# bio 
+from Bio import SeqIO
+import itertools
+import string
+
 # sidechainnet
 
 from sidechainnet.utils.sequence import ProteinVocabulary, ONE_TO_THREE_LETTER_MAP
@@ -214,9 +219,25 @@ def coords2pdb(seq, coords, cloud_mask, prefix="", name="af2_struct.pdb"):
     pred.to_pdb(prefix+name)
 
 
-# sidechainnet / other data utils
+# adapted from https://github.com/facebookresearch/esm
 
-def get_esm_embedd(seq, embedd_model, batch_converter, embedd_type="per_tok"):
+def remove_insertions(sequence: str) -> str:
+    """ Removes any insertions into the sequence. Needed to load aligned sequences in an MSA. """
+    deletekeys = dict.fromkeys(string.ascii_lowercase)
+    deletekeys["."] = None
+    deletekeys["*"] = None
+    translation = str.maketrans(deletekeys)
+    return sequence.translate(translation)
+
+def read_msa(filename: str, nseq: int):
+    """ Reads the first nseq sequences from an MSA file, automatically removes insertions."""
+    return [(record.description, remove_insertions(str(record.seq)))
+            for record in itertools.islice(SeqIO.parse(filename, "fasta"), nseq)]
+
+
+# sidechainnet / MSA / other data utils
+
+def get_esm_embedd(seq, embedd_model, batch_converter, msa_data=None, embedd_type="per_tok"):
     """ Returns the ESM embeddings for a protein. 
         Inputs: 
         * seq: (L,) tensor of ints (in sidechainnet int-char convention)
@@ -224,13 +245,29 @@ def get_esm_embedd(seq, embedd_model, batch_converter, embedd_type="per_tok"):
         * batch_converter: ESM batch converter (see train_end2end.py for an example)
         * embedd_type: one of ["mean", "per_tok"]. 
                        "per_tok" is recommended if working with sequences.
+        Outputs: tensor of (batch, n_seqs, L, embedd_dim)
+            * n_seqs: number of sequences in the MSA. 1 for ESM-1b
+            * embedd_dim: number of embedding dimensions. 
+                          768 for MSA_Transformer and 1280 for ESM-1b
     """
     str_seq = "".join([VOCAB._int2char[x]for x in seq.cpu().numpy()])
-    batch_labels, batch_strs, batch_tokens = batch_converter( [(0, str_seq)] )
-    with torch.no_grad():
-        results = embedd_model(batch_tokens, repr_layers=[33], return_contacts=False)
-    # index 0 is for start token. so take from 1 one
-    token_reps = results["representations"][33][ 1 : len(str_seq) + 1].to(seq.device)
+    # use MSA transformer
+    if msa_data is not None: 
+        msa_batch_labels, msa_batch_strs, msa_batch_tokens = batch_converter(msa_data)
+        with torch.no_grad():
+            results = embedd_model(msa_batch_tokens.to(seq.device), repr_layers=[12], return_contacts=False)
+        # index 0 is for start token. so take from 1 one
+        token_reps = results["representations"][12][0, :,  1 : len(str_seq) + 1]
+        
+    # base ESM case
+    else: 
+        batch_labels, batch_strs, batch_tokens = batch_converter( [(0, str_seq)] )
+        with torch.no_grad():
+            results = embedd_model(batch_tokens.to(seq.device), repr_layers=[33], return_contacts=False)
+        # index 0 is for start token. so take from 1 one
+        token_reps = results["representations"][33][:, 1 : len(str_seq) + 1].unsqueeze(dim=1)
+        
+    
     if embedd_type == "mean":
         token_reps = token_reps.mean(dim=0)
     return token_reps
@@ -467,7 +504,8 @@ def center_distogram_torch(distogram, bins=DISTANCE_THRESHOLDS, min_t=1., center
         dispersion = torch.zeros_like(central, device=device)
     # rescale to 0-1. lower std / var  --> weight=1. set potential nan's to 0
     weights = mask / (1+dispersion)
-    weights[weights != weights] = 0.
+    weights[weights != weights] *= 0.
+    weights[:, diag_idxs, diag_idxs] *= 0.
     return central, weights
 
 # distance matrix to 3d coords: https://github.com/scikit-learn/scikit-learn/blob/42aff4e2e/sklearn/manifold/_mds.py#L279
@@ -490,7 +528,7 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     # start
     batch, N, _ = pre_dist_mat.shape
     diag_idxs = np.arange(N)
-    his = [torch.tensor([np.inf])]
+    his = [torch.tensor([np.inf]*batch)]
     # init random coords
     best_stress = float("Inf") * torch.ones(batch, device = device).type(dtype)
     best_3d_coords = 2*torch.rand(batch, N, 3, device = device).type(dtype) - 1
@@ -500,7 +538,7 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
         dist_mat = torch.cdist(best_3d_coords, best_3d_coords, p=2).clone()
         stress   = ( weights * (dist_mat - pre_dist_mat)**2 ).sum(dim=(-1,-2)) * 0.5
         # perturb - update X using the Guttman transform - sklearn-like
-        dist_mat[ dist_mat == 0 ] = 1e-7
+        dist_mat[ dist_mat <= 0 ] += 1e-7
         ratio = weights * (pre_dist_mat / dist_mat)
         B = -ratio
         B[:, diag_idxs, diag_idxs] += ratio.sum(dim=-1)
@@ -518,10 +556,9 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
             break
 
         best_3d_coords = coords
-        best_stress = (stress / dis)
-        his.append(best_stress)
+        his.append( stress / dis )
 
-    return torch.transpose(best_3d_coords, -1,-2), torch.cat(his)
+    return torch.transpose(best_3d_coords, -1,-2), torch.stack(his)
 
 def mds_numpy(pre_dist_mat, weights=None, iters=10, tol=1e-5, verbose=2):
     """ Gets distance matrix. Outputs 3d. See below for wrapper. 
@@ -739,22 +776,31 @@ def kabsch_numpy(X, Y):
 
 # metrics - more formulas here: http://predictioncenter.org/casp12/doc/help.html
 
-def distmat_loss_torch(X, Y, p=2, q=2, distmat_mask=None):
+def distmat_loss_torch(X=None, Y=None, X_mat=None, Y_mat=None, p=2, q=2, distmat_mask=None):
     """ Calculates a loss on the distance matrix - no need to align structs.
         Inputs: 
-        * X: (N, d) tensor. the predicted stricture 
-        * Y: (N, d) tensor. the true structure
+        * X: (N, d) tensor. the predicted structure. One of (X, X_mat) is needed.
+        * X_mat: (N, N) tensor. the predicted distance matrix. Optional ()
+        * Y: (N, d) tensor. the true structure. One of (Y, Y_mat) is needed.
+        * Y_mat: (N, N) tensor. the predicted distance matrix. Optional ()
         * p: int. power for the distance calculation (2 for euclidean)
         * q: float. power for the scaling of the loss (2 for MSE, 1 for MAE, etc)
         * distmat_mask: (N, N) mask (boolean or weights for each ij pos). optional.
     """
+    assert (X is not None or X_mat is not None) and \
+           (Y is not None or Y_mat is not None), "The true and predicted coords or dist mats must be provided"
+    # calculate distance matrices
+    if X_mat is None: 
+        X_mat = torch.cdist(X, X, p=p)
+    if Y_mat is None: 
+        Y_mat = torch.cdist(Y, Y, p=p)
+    if distmat_mask is None:
+        distmat_mask = torch.ones_like(Y_mat).bool()
     # **2 ensures always positive. Later scale back to desired power
-    loss = ( torch.cdist(X, X, p=p) - torch.cdist(Y, Y, p=p) )**2 
+    loss = ( X_mat - Y_mat )**2 
     if q != 2:
         loss = loss**(q/2)
-    if distmat_mask is not None:
-        loss *= distmat_mask.float()
-    return loss.mean()
+    return loss[distmat_mask].mean()
 
 def rmsd_torch(X, Y):
     """ Assumes x,y are both (B x D x N). See below for wrapper. """
